@@ -20,13 +20,11 @@
  * ----------------------------------------------------------------------- */
 
 #include <dirent.h>
-#include <fcntl.h>
 #include <getopt.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
 #include <sys/ioctl.h>
 #include <ctype.h>
 #include <sys/types.h>
@@ -42,6 +40,7 @@
 #include "automount.h"
 #ifdef LIBXML2_WORKAROUND
 #include <dlfcn.h>
+#include <libxml/parser.h>
 #endif
 
 const char *program;		/* Initialized with argv[0] */
@@ -59,17 +58,24 @@ static char *pid_file = NULL;		/* File in which to keep pid */
 unsigned int global_random_selection;	/* use random policy when selecting
 					 * which multi-mount host to mount */
 long global_negative_timeout = -1;
+int do_force_unlink = 0;		/* Forceably unlink mount tree at startup */
 
 static int start_pipefd[2];
-static int st_stat = 0;
+static int st_stat = 1;
 static int *pst_stat = &st_stat;
 static pthread_t state_mach_thid;
 
+static sigset_t block_sigs;
+
 /* Pre-calculated kernel packet length */
 static size_t kpkt_len;
 
-/* Attribute to create detached thread */
-pthread_attr_t thread_attr;
+/* Does kernel know about SOCK_CLOEXEC and friends */
+static int cloexec_works = 0;
+
+/* Attributes for creating detached and joinable threads */
+pthread_attr_t th_attr;
+pthread_attr_t th_attr_detached;
 
 struct master_readmap_cond mrc = {
 	PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, 0, NULL, 0, 0, 0, 0};
@@ -81,9 +87,10 @@ pthread_key_t key_thread_stdenv_vars;
 
 #define MAX_OPEN_FILES		10240
 
+int aquire_flag_file(void);
+void release_flag_file(void);
 static int umount_all(struct autofs_point *ap, int force);
 
-extern pthread_mutex_t master_mutex;
 extern struct master *master_list;
 
 static int do_mkdir(const char *parent, const char *path, mode_t mode)
@@ -123,8 +130,8 @@ static int do_mkdir(const char *parent, const char *path, mode_t mode)
 
 int mkdir_path(const char *path, mode_t mode)
 {
-	char *buf = alloca(strlen(path) + 1);
-	char *parent = alloca(strlen(path) + 1);
+	char buf[PATH_MAX];
+	char parent[PATH_MAX];
 	const char *cp = path, *lcp = path;
 	char *bp = buf, *pp = parent;
 
@@ -159,7 +166,7 @@ int mkdir_path(const char *path, mode_t mode)
 int rmdir_path(struct autofs_point *ap, const char *path, dev_t dev)
 {
 	int len = strlen(path);
-	char *buf = alloca(len + 1);
+	char buf[PATH_MAX];
 	char *cp;
 	int first = 1;
 	struct stat st;
@@ -245,9 +252,17 @@ static int walk_tree(const char *base, int (*fn) (unsigned logopt,
 						  int, void *), int incl, unsigned logopt, void *arg)
 {
 	char buf[PATH_MAX + 1];
-	struct stat st;
+	struct stat st, *pst = &st;
+	int ret;
 
-	if (lstat(base, &st) != -1 && (fn) (logopt, base, &st, 0, arg)) {
+	if (!is_mounted(_PATH_MOUNTED, base, MNTS_REAL))
+		ret = lstat(base, pst);
+	else {
+		pst = NULL;
+		ret = 0;
+	}
+
+	if (ret != -1 && (fn) (logopt, base, pst, 0, arg)) {
 		if (S_ISDIR(st.st_mode)) {
 			struct dirent **de;
 			int n;
@@ -281,7 +296,7 @@ static int walk_tree(const char *base, int (*fn) (unsigned logopt,
 			free(de);
 		}
 		if (incl)
-			(fn) (logopt, base, &st, 1, arg);
+			(fn) (logopt, base, pst, 1, arg);
 	}
 	return 0;
 }
@@ -292,6 +307,9 @@ static int rm_unwanted_fn(unsigned logopt, const char *file, const struct stat *
 	char buf[MAX_ERR_BUF];
 	struct stat newst;
 
+	if (!st)
+		return 0;
+
 	if (when == 0) {
 		if (st->st_dev != dev)
 			return 0;
@@ -342,8 +360,8 @@ static int counter_fn(unsigned logopt, const char *file, const struct stat *st,
 {
 	struct counter_args *counter = (struct counter_args *) arg;
 
-	if (S_ISLNK(st->st_mode) || (S_ISDIR(st->st_mode) 
-		&& st->st_dev != counter->dev)) {
+	if (!st || (S_ISLNK(st->st_mode) || (S_ISDIR(st->st_mode)
+		&& st->st_dev != counter->dev))) {
 		counter->count++;
 		return 0;
 	}
@@ -367,12 +385,24 @@ int count_mounts(unsigned logopt, const char *path, dev_t dev)
 
 static void check_rm_dirs(struct autofs_point *ap, const char *path, int incl)
 {
-	if ((!ap->ghost) ||
+	/*
+	 * If we're a submount the kernel can't know we're trying to
+	 * shutdown and so cannot block processes walking into the
+	 * mount point directory. If this is the call to umount_multi()
+	 * made during shutdown (incl == 0) we have to leave any mount
+	 * point directories in place so we can recover if needed. The
+	 * umount itself will clean these directories up for us
+	 * automagically.
+	 */
+	if (!incl && ap->submount)
+		return;
+
+	if ((!(ap->flags & MOUNT_FLAG_GHOST)) ||
 	    (ap->state == ST_SHUTDOWN_PENDING ||
 	     ap->state == ST_SHUTDOWN_FORCE ||
 	     ap->state == ST_SHUTDOWN))
 		rm_unwanted(ap->logopt, path, incl, ap->dev);
-	else if (ap->ghost && (ap->type == LKP_INDIRECT))
+	else if ((ap->flags & MOUNT_FLAG_GHOST) && (ap->type == LKP_INDIRECT))
 		rm_unwanted(ap->logopt, path, 0, ap->dev);
 }
 
@@ -388,8 +418,6 @@ static void update_map_cache(struct autofs_point *ap, const char *path)
 	else
 		key = path;
 
-	pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
-	master_source_readlock(ap->entry);
 	map = ap->entry->maps;
 	while (map) {
 		struct mapent *me = NULL;
@@ -411,7 +439,6 @@ static void update_map_cache(struct autofs_point *ap, const char *path)
 
 		map = map->next;
 	}
-	pthread_cleanup_pop(1);
 
 	return;
 }
@@ -444,20 +471,17 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 	pthread_cleanup_push(cache_lock_cleanup, mc);
 
 	if (me->multi) {
-		char *root, *base;
-		size_t ap_len;
+		char root[PATH_MAX];
+		char *base;
 		int cur_state;
 
-		ap_len = strlen(ap->path);
-
-		if (!strchr(me->multi->key, '/')) {
+		if (!strchr(me->multi->key, '/'))
 			/* Indirect multi-mount root */
-			root = alloca(ap_len + strlen(me->multi->key) + 2);
-			strcpy(root, ap->path);
-			strcat(root, "/");
-			strcat(root, me->multi->key);
-		} else
-			root = me->multi->key;
+			/* sprintf okay - if it's mounted, it's
+			 * PATH_MAX or less bytes */
+			sprintf(root, "%s/%s", ap->path, me->multi->key);
+		else
+			strcpy(root, me->multi->key);
 
 		if (is_mm_root)
 			base = NULL;
@@ -466,8 +490,8 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 		/* Lock the closest parent nesting point for umount */
-		cache_multi_lock(me->parent);
-		if (umount_multi_triggers(ap, root, me, base)) {
+		cache_multi_writelock(me->parent);
+		if (umount_multi_triggers(ap, me, root, base)) {
 			warn(ap->logopt,
 			     "some offset mounts still present under %s", path);
 			left++;
@@ -501,9 +525,8 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 int umount_multi(struct autofs_point *ap, const char *path, int incl)
 {
 	struct mapent_cache *nc;
-	struct statfs fs;
 	int is_autofs_fs;
-	int ret, left;
+	int left;
 
 	debug(ap->logopt, "path %s incl %d", path, incl);
 
@@ -515,13 +538,9 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 	}
 	cache_unlock(nc);
 
-	ret = statfs(path, &fs);
-	if (ret == -1) {
-		error(ap->logopt, "could not stat fs of %s", path);
-		return 1;
-	}
-
-	is_autofs_fs = fs.f_type == (__SWORD_TYPE) AUTOFS_SUPER_MAGIC ? 1 : 0;
+	is_autofs_fs = 0;
+	if (master_find_submount(ap, path))
+		is_autofs_fs = 1;
 
 	left = 0;
 
@@ -555,7 +574,7 @@ static int umount_all(struct autofs_point *ap, int force)
 	return left;
 }
 
-int umount_autofs(struct autofs_point *ap, int force)
+int umount_autofs(struct autofs_point *ap, const char *root, int force)
 {
 	int ret = 0;
 
@@ -572,47 +591,13 @@ int umount_autofs(struct autofs_point *ap, int force)
 	if (ap->type == LKP_INDIRECT) {
 		if (umount_all(ap, force) && !force)
 			return -1;
-		ret = umount_autofs_indirect(ap);
+		ret = umount_autofs_indirect(ap, root);
 	} else
 		ret = umount_autofs_direct(ap);
 
 	return ret;
 }
 
-int send_ready(unsigned logopt, int ioctlfd, unsigned int wait_queue_token)
-{
-	char buf[MAX_ERR_BUF];
-
-	if (wait_queue_token == 0)
-		return 0;
-
-	debug(logopt, "token = %d", wait_queue_token);
-
-	if (ioctl(ioctlfd, AUTOFS_IOC_READY, wait_queue_token) < 0) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr("AUTOFS_IOC_READY: error %s", estr);
-		return 1;
-	}
-	return 0;
-}
-
-int send_fail(unsigned logopt, int ioctlfd, unsigned int wait_queue_token)
-{
-	char buf[MAX_ERR_BUF];
-
-	if (wait_queue_token == 0)
-		return 0;
-
-	debug(logopt, "token = %d", wait_queue_token);
-
-	if (ioctl(ioctlfd, AUTOFS_IOC_FAIL, wait_queue_token) < 0) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr("AUTOFS_IOC_FAIL: error %s", estr);
-		return 1;
-	}
-	return 0;
-}
-
 static size_t get_kpkt_len(void)
 {
 	size_t pkt_len = sizeof(struct autofs_v5_packet);
@@ -689,7 +674,7 @@ static char *automount_path_to_fifo(unsigned logopt, const char *path)
 static int create_logpri_fifo(struct autofs_point *ap)
 {
 	int ret = -1;
-	int fd, cl_flags;
+	int fd;
 	char *fifo_name;
 	char buf[MAX_ERR_BUF];
 
@@ -715,7 +700,7 @@ static int create_logpri_fifo(struct autofs_point *ap)
 		goto out_free;
 	}
 
-	fd = open(fifo_name, O_RDWR|O_NONBLOCK);
+	fd = open_fd(fifo_name, O_RDWR|O_NONBLOCK);
 	if (fd < 0) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		crit(ap->logopt,
@@ -725,11 +710,6 @@ static int create_logpri_fifo(struct autofs_point *ap)
 		goto out_free;
 	}
 
-	if ((cl_flags = fcntl(fd, F_GETFD, 0)) != -1) {
-		cl_flags |= FD_CLOEXEC;
-		fcntl(fd, F_SETFD, cl_flags);
-	}
-
 	ap->logpri_fifo = fd;
 
 out_free:
@@ -737,7 +717,7 @@ out_free:
 	return ret;
 }
 
-static int destroy_logpri_fifo(struct autofs_point *ap)
+int destroy_logpri_fifo(struct autofs_point *ap)
 {
 	int ret = -1;
 	int fd = ap->logpri_fifo;
@@ -890,7 +870,7 @@ static int set_log_priority(const char *path, int priority)
 	return 0;
 }
 
-static int get_pkt(struct autofs_point *ap, union autofs_packet_union *pkt)
+static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 {
 	struct pollfd fds[3];
 	int pollfds = 3;
@@ -916,38 +896,22 @@ static int get_pkt(struct autofs_point *ap, union autofs_packet_union *pkt)
 		}
 
 		if (fds[1].revents & POLLIN) {
-			enum states next_state, post_state;
+			enum states next_state;
 			size_t read_size = sizeof(next_state);
 			int state_pipe;
 
-			next_state = post_state = ST_INVAL;
+			next_state = ST_INVAL;
 
-			state_mutex_lock(ap);
+			st_mutex_lock();
 
 			state_pipe = ap->state_pipe[0];
 
 			if (fullread(state_pipe, &next_state, read_size)) {
-				state_mutex_unlock(ap);
+				st_mutex_unlock();
 				continue;
 			}
 
-			if (next_state != ST_INVAL && next_state != ap->state) {
-				if (next_state != ST_SHUTDOWN)
-					post_state = next_state;
-				else
-					ap->state = ST_SHUTDOWN;
-			}
-
-			state_mutex_unlock(ap);
-
-			if (post_state != ST_INVAL) {
-				if (post_state == ST_SHUTDOWN_PENDING ||
-				    post_state == ST_SHUTDOWN_FORCE) {
-					alarm_delete(ap);
-					st_remove_tasks(ap);
-				}
-				st_add_task(ap, post_state);
-			}
+			st_mutex_unlock();
 
 			if (next_state == ST_SHUTDOWN)
 				return -1;
@@ -965,14 +929,14 @@ static int get_pkt(struct autofs_point *ap, union autofs_packet_union *pkt)
 
 int do_expire(struct autofs_point *ap, const char *name, int namelen)
 {
-	char buf[PATH_MAX + 1];
+	char buf[PATH_MAX];
 	int len, ret;
 
 	if (*name != '/') {
 		len = ncat_path(buf, sizeof(buf), ap->path, name, namelen);
 	} else {
 		len = snprintf(buf, PATH_MAX, "%s", name);
-		if (len > PATH_MAX)
+		if (len >= PATH_MAX)
 			len = 0;
 	}
 
@@ -983,18 +947,21 @@ int do_expire(struct autofs_point *ap, const char *name, int namelen)
 
 	info(ap->logopt, "expiring path %s", buf);
 
+	pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
+	master_source_readlock(ap->entry);
 	ret = umount_multi(ap, buf, 1);
 	if (ret == 0)
 		info(ap->logopt, "expired %s", buf);
 	else
 		warn(ap->logopt, "couldn't complete expire of %s", buf);
+	pthread_cleanup_pop(1);
 
 	return ret;
 }
 
 static int autofs_init_ap(struct autofs_point *ap)
 {
-	int pipefd[2], cl_flags;
+	int pipefd[2];
 
 	if ((ap->state != ST_INIT)) {
 		/* This can happen if an autofs process is already running*/
@@ -1005,7 +972,7 @@ static int autofs_init_ap(struct autofs_point *ap)
 	ap->pipefd = ap->kpipefd = ap->ioctlfd = -1;
 
 	/* Pipe for kernel communications */
-	if (pipe(pipefd) < 0) {
+	if (open_pipe(pipefd) < 0) {
 		crit(ap->logopt,
 		     "failed to create commumication pipe for autofs path %s",
 		     ap->path);
@@ -1015,18 +982,8 @@ static int autofs_init_ap(struct autofs_point *ap)
 	ap->pipefd = pipefd[0];
 	ap->kpipefd = pipefd[1];
 
-	if ((cl_flags = fcntl(ap->pipefd, F_GETFD, 0)) != -1) {
-		cl_flags |= FD_CLOEXEC;
-		fcntl(ap->pipefd, F_SETFD, cl_flags);
-	}
-
-	if ((cl_flags = fcntl(ap->kpipefd, F_GETFD, 0)) != -1) {
-		cl_flags |= FD_CLOEXEC;
-		fcntl(ap->kpipefd, F_SETFD, cl_flags);
-	}
-
 	/* Pipe state changes from signal handler to main loop */
-	if (pipe(ap->state_pipe) < 0) {
+	if (open_pipe(ap->state_pipe) < 0) {
 		crit(ap->logopt,
 		     "failed create state pipe for autofs path %s", ap->path);
 		close(ap->pipefd);
@@ -1034,16 +991,6 @@ static int autofs_init_ap(struct autofs_point *ap)
 		return -1;
 	}
 
-	if ((cl_flags = fcntl(ap->state_pipe[0], F_GETFD, 0)) != -1) {
-		cl_flags |= FD_CLOEXEC;
-		fcntl(ap->state_pipe[0], F_SETFD, cl_flags);
-	}
-
-	if ((cl_flags = fcntl(ap->state_pipe[1], F_GETFD, 0)) != -1) {
-		cl_flags |= FD_CLOEXEC;
-		fcntl(ap->state_pipe[1], F_SETFD, cl_flags);
-	}
-
 	if (create_logpri_fifo(ap) < 0) {
 		logmsg("could not create FIFO for path %s\n", ap->path);
 		logmsg("dynamic log level changes not available for %s", ap->path);
@@ -1052,7 +999,7 @@ static int autofs_init_ap(struct autofs_point *ap)
 	return 0;
 }
 
-static int mount_autofs(struct autofs_point *ap)
+static int mount_autofs(struct autofs_point *ap, const char *root)
 {
 	int status = 0;
 
@@ -1062,19 +1009,19 @@ static int mount_autofs(struct autofs_point *ap)
 	if (ap->type == LKP_DIRECT)
 		status = mount_autofs_direct(ap);
 	else
-		status = mount_autofs_indirect(ap);
+		status = mount_autofs_indirect(ap, root);
 
 	if (status < 0)
 		return -1;
 
-	ap->state = ST_READY;
+	st_add_task(ap, ST_READY);
 
 	return 0;
 }
 
 static int handle_packet(struct autofs_point *ap)
 {
-	union autofs_packet_union pkt;
+	union autofs_v5_packet_union pkt;
 
 	if (get_pkt(ap, &pkt))
 		return -1;
@@ -1098,10 +1045,11 @@ static int handle_packet(struct autofs_point *ap)
 	return -1;
 }
 
-static void become_daemon(unsigned foreground)
+static void become_daemon(unsigned foreground, unsigned daemon_check)
 {
 	FILE *pidfp;
 	char buf[MAX_ERR_BUF];
+	int res;
 	pid_t pid;
 
 	/* Don't BUSY any directories unnecessarily */
@@ -1111,22 +1059,26 @@ static void become_daemon(unsigned foreground)
 		exit(0);
 	}
 
-	if (pipe(start_pipefd) < 0) {
+	if (open_pipe(start_pipefd) < 0) {
 		fprintf(stderr, "%s: failed to create start_pipefd.\n",
 			program);
 		exit(0);
 	}
 
 	/* Detach from foreground process */
-	if (foreground)
+	if (foreground) {
+		if (daemon_check && !aquire_flag_file()) {
+			fprintf(stderr, "%s: program is already running.\n",
+				program);
+			exit(1);
+		}
 		log_to_stderr();
-	else {
+	} else {
 		pid = fork();
 		if (pid > 0) {
-			int r;
 			close(start_pipefd[1]);
-			r = read(start_pipefd[0], pst_stat, sizeof(*pst_stat));
-			if (r < 0)
+			res = read(start_pipefd[0], pst_stat, sizeof(*pst_stat));
+			if (res < 0)
 				exit(1);
 			exit(*pst_stat);
 		} else if (pid < 0) {
@@ -1136,6 +1088,18 @@ static void become_daemon(unsigned foreground)
 		}
 		close(start_pipefd[0]);
 
+		if (daemon_check && !aquire_flag_file()) {
+			fprintf(stderr, "%s: program is already running.\n",
+				program);
+			/* Return success if already running */
+			st_stat = 0;
+			res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+			if (res < 0)
+				exit(1);
+			close(start_pipefd[1]);
+			exit(*pst_stat);
+		}
+
 		/*
 		 * Make our own process group for "magic" reason: processes that share
 		 * our pgrp see the raw filesystem behind the magic.
@@ -1143,7 +1107,9 @@ static void become_daemon(unsigned foreground)
 		if (setsid() == -1) {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			fprintf(stderr, "setsid: %s", estr);
-			exit(1);
+			res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+			close(start_pipefd[1]);
+			exit(*pst_stat);
 		}
 		log_to_syslog();
 	}
@@ -1234,7 +1200,7 @@ static pthread_t do_signals(struct master *master, int sig)
 	if (status)
 		fatal(status);
 
-	status = pthread_create(&thid, &thread_attr, do_notify_state, &r_sig);
+	status = pthread_create(&thid, &th_attr_detached, do_notify_state, &r_sig);
 	if (status) {
 		error(master->logopt,
 		      "mount state notify thread create failed");
@@ -1323,7 +1289,7 @@ static int do_hup_signal(struct master *master, time_t age)
 
 	master->reading = 1;
 
-	status = pthread_create(&thid, &thread_attr, do_read_master, NULL);
+	status = pthread_create(&thid, &th_attr_detached, do_read_master, NULL);
 	if (status) {
 		error(logopt,
 		      "master read map thread create failed");
@@ -1358,7 +1324,7 @@ static void *statemachine(void *arg)
 	sigset_t signalset;
 	int sig;
 
-	sigfillset(&signalset);
+	memcpy(&signalset, &block_sigs, sizeof(signalset));
 	sigdelset(&signalset, SIGCHLD);
 	sigdelset(&signalset, SIGCONT);
 
@@ -1369,8 +1335,22 @@ static void *statemachine(void *arg)
 		case SIGTERM:
 		case SIGINT:
 		case SIGUSR2:
-			if (master_list_empty(master_list))
-				return NULL;
+			master_mutex_lock();
+			if (list_empty(&master_list->completed)) {
+				if (list_empty(&master_list->mounts)) {
+					master_mutex_unlock();
+					return NULL;
+				}
+			} else {
+				if (master_done(master_list)) {
+					master_mutex_unlock();
+					return NULL;
+				}
+				master_mutex_unlock();
+				break;
+			}
+			master_mutex_unlock();
+
 		case SIGUSR1:
 			do_signals(master_list, sig);
 			break;
@@ -1408,41 +1388,52 @@ static void return_start_status(void *arg)
 		fatal(status);
 }
 
-static void mutex_operation_wait(pthread_mutex_t *mutex)
+int handle_mounts_startup_cond_init(struct startup_cond *suc)
 {
 	int status;
 
-	/*
-	 * Unlock a mutex, but wait for a pending operation
-	 * if one is in progress
-	 */
-	status = pthread_mutex_trylock(mutex);
+	status = pthread_mutex_init(&suc->mutex, NULL);
+	if (status)
+		return status;
+
+	status = pthread_cond_init(&suc->cond, NULL);
 	if (status) {
-		if (status == EBUSY) {
-			/* Mutex locked - do we own it */
-			status = pthread_mutex_unlock(mutex);
-			if (status) {
-				if (status != EPERM)
-					fatal(status);
-			} else
-				return;
-
-			status = pthread_mutex_lock(mutex);
-			if (status)
-				fatal(status);
-		} else
-			fatal(status);
-
-		/* Operation complete, release it */
-		status = pthread_mutex_unlock(mutex);
+		status = pthread_mutex_destroy(&suc->mutex);
 		if (status)
 			fatal(status);
-	} else {
-		status = pthread_mutex_unlock(mutex);
+		return status;
+	}
+
+	status = pthread_mutex_lock(&suc->mutex);
+	if (status) {
+		status = pthread_mutex_destroy(&suc->mutex);
+		if (status)
+			fatal(status);
+		status = pthread_cond_destroy(&suc->cond);
 		if (status)
 			fatal(status);
 	}
 
+	return 0;
+}
+
+void handle_mounts_startup_cond_destroy(void *arg)
+{
+	struct startup_cond *suc = (struct startup_cond *) arg;
+	int status;
+
+	status = pthread_mutex_unlock(&suc->mutex);
+	if (status)
+		fatal(status);
+
+	status = pthread_mutex_destroy(&suc->mutex);
+	if (status)
+		fatal(status);
+
+	status = pthread_cond_destroy(&suc->cond);
+	if (status)
+		fatal(status);
+
 	return;
 }
 
@@ -1459,24 +1450,33 @@ static void handle_mounts_cleanup(void *arg)
 	submount = ap->submount;
 
 	strcpy(path, ap->path);
-	if (!submount && strcmp(ap->path, "/-") && ap->dir_created)
+	if (!submount && strcmp(ap->path, "/-") &&
+	    ap->flags & MOUNT_FLAG_DIR_CREATED)
 		clean = 1;
 
-	/* If we have been canceled then we may hold the state mutex. */
-	mutex_operation_wait(&ap->state_mutex);
+	if (submount) {
+		/* We are finishing up */
+		ap->parent->submnt_count--;
+		list_del_init(&ap->mounts);
+	}
 
-	alarm_delete(ap);
-	st_remove_tasks(ap);
-
-	umount_autofs(ap, 1);
+	master_remove_mapent(ap->entry);
+	master_source_unlock(ap->entry);
 
 	destroy_logpri_fifo(ap);
-	master_signal_submount(ap, MASTER_SUBMNT_JOIN);
-	master_remove_mapent(ap->entry);
-	master_free_mapent_sources(ap->entry, 1);
-	master_free_mapent(ap->entry);
 
-	sched_yield();
+	/*
+	 * Submounts are detached threads and don't belong to the
+	 * master map entry list so we need to free their resources
+	 * here.
+	 */
+	if (submount) {
+		mounts_mutex_unlock(ap->parent);
+		master_source_unlock(ap->parent->entry);
+		master_free_mapent_sources(ap->entry, 1);
+		master_free_mapent(ap->entry);
+	}
+	master_mutex_unlock();
 
 	if (clean) {
 		if (rmdir(path) == -1) {
@@ -1488,8 +1488,12 @@ static void handle_mounts_cleanup(void *arg)
 
 	info(logopt, "shut down path %s", path);
 
-	/* If we are the last tell the state machine to shutdown */
-	if (!submount && master_list_empty(master_list))
+	/*
+	 * If we are not a submount send a signal to the signal handler
+	 * so it can join with any completed handle_mounts() threads and
+	 * perform final cleanup.
+	 */
+	if (!submount)
 		pthread_kill(state_mach_thid, SIGTERM);
 	
 	return;
@@ -1497,35 +1501,47 @@ static void handle_mounts_cleanup(void *arg)
 
 void *handle_mounts(void *arg)
 {
+	struct startup_cond *suc;
 	struct autofs_point *ap;
 	int cancel_state, status = 0;
+	char *root;
 
-	ap = (struct autofs_point *) arg;
+	suc = (struct startup_cond *) arg;
 
-	pthread_cleanup_push(return_start_status, &suc);
+	ap = suc->ap;
+	root = strdup(suc->root);
+
+	pthread_cleanup_push(return_start_status, suc);
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel_state);
 
-	state_mutex_lock(ap);
-
-	status = pthread_mutex_lock(&suc.mutex);
+	status = pthread_mutex_lock(&suc->mutex);
 	if (status) {
 		logerr("failed to lock startup condition mutex!");
 		fatal(status);
 	}
 
-	if (mount_autofs(ap) < 0) {
-		crit(ap->logopt, "mount of %s failed!", ap->path);
-		suc.status = 1;
-		state_mutex_unlock(ap);
-		umount_autofs(ap, 1);
+	if (!root) {
+		crit(ap->logopt, "failed to alloc string root");
+		suc->status = 1;
 		pthread_setcancelstate(cancel_state, NULL);
 		pthread_exit(NULL);
 	}
 
-	if (ap->ghost && ap->type != LKP_DIRECT)
+	if (mount_autofs(ap, root) < 0) {
+		crit(ap->logopt, "mount of %s failed!", ap->path);
+		suc->status = 1;
+		umount_autofs(ap, root, 1);
+		free(root);
+		pthread_setcancelstate(cancel_state, NULL);
+		pthread_exit(NULL);
+	}
+
+	free(root);
+
+	if (ap->flags & MOUNT_FLAG_GHOST && ap->type != LKP_DIRECT)
 		info(ap->logopt, "ghosting enabled");
 
-	suc.status = 0;
+	suc->status = 0;
 	pthread_cleanup_pop(1);
 
 	/* We often start several automounters at the same time.  Add some
@@ -1533,56 +1549,70 @@ void *handle_mounts(void *arg)
 	if (!ap->submount && ap->exp_timeout)
 		alarm_add(ap, ap->exp_runfreq + rand() % ap->exp_runfreq);
 
-	pthread_cleanup_push(handle_mounts_cleanup, ap);
 	pthread_setcancelstate(cancel_state, NULL);
 
-	state_mutex_unlock(ap);
-
 	while (ap->state != ST_SHUTDOWN) {
 		if (handle_packet(ap)) {
-			int ret, result;
+			int ret, cur_state;
+
+			/*
+			 * If we're a submount we need to ensure our parent
+			 * doesn't try to mount us again until our shutdown
+			 * is complete and that any outstanding mounts are
+			 * completed before we try to shutdown.
+			 */
+			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+
+			master_mutex_lock();
+
+			if (ap->submount) {
+				master_source_writelock(ap->parent->entry);
+				mounts_mutex_lock(ap->parent);
+			}
+
+			master_source_writelock(ap->entry);
+
+			if (ap->state != ST_SHUTDOWN) {
+				if (!ap->submount)
+					alarm_add(ap, ap->exp_runfreq);
+				/* Return to ST_READY is done immediately */
+				st_add_task(ap, ST_READY);
+				master_source_unlock(ap->entry);
+				if (ap->submount) {
+					mounts_mutex_unlock(ap->parent);
+					master_source_unlock(ap->parent->entry);
+				}
+
+				master_mutex_unlock();
+
+				pthread_setcancelstate(cur_state, NULL);
+				continue;
+			}
+
+			alarm_delete(ap);
+			st_remove_tasks(ap);
+			st_wait_task(ap, ST_ANY, 0);
 
-			state_mutex_lock(ap);
 			/*
 			 * For a direct mount map all mounts have already gone
-			 * by the time we get here.
+			 * by the time we get here and since we only ever
+			 * umount direct mounts at shutdown there is no need
+			 * to check for possible recovery.
 			 */
 			if (ap->type == LKP_DIRECT) {
-				status = 1;
-				state_mutex_unlock(ap);
+				umount_autofs(ap, NULL, 1);
 				break;
 			}
 
 			/*
-			 * If the ioctl fails assume the kernel doesn't have
-			 * AUTOFS_IOC_ASKUMOUNT and just continue.
+			 * If umount_autofs returns non-zero it wasn't able
+			 * to complete the umount and has left the mount intact
+			 * so we can continue. This can happen if a lookup
+			 * occurs while we're trying to umount.
 			 */
-			ret = ioctl(ap->ioctlfd, AUTOFS_IOC_ASKUMOUNT, &result);
-			if (ret == -1) {
-				state_mutex_unlock(ap);
+			ret = umount_autofs(ap, NULL, 1);
+			if (!ret)
 				break;
-			}
-
-			/* OK to exit */
-			if (ap->state == ST_SHUTDOWN) {
-				if (result) {
-					state_mutex_unlock(ap);
-					break;
-				}
-#ifdef ENABLE_IGNORE_BUSY_MOUNTS
-				/*
-				 * There weren't any active mounts but if the
-				 * filesystem is busy there may be a mount
-				 * request in progress so return to the ready
-				 * state unless a shutdown has been explicitly
-				 * requested.
-				 */
-				if (ap->shutdown) {
-					state_mutex_unlock(ap);
-					break;
-				}
-#endif
-			}
 
 			/* Failed shutdown returns to ready */
 			warn(ap->logopt,
@@ -1590,14 +1620,22 @@ void *handle_mounts(void *arg)
 			     ap->path);
 			if (!ap->submount)
 				alarm_add(ap, ap->exp_runfreq);
-			nextstate(ap->state_pipe[1], ST_READY);
+			/* Return to ST_READY is done immediately */
+			st_add_task(ap, ST_READY);
+			master_source_unlock(ap->entry);
+			if (ap->submount) {
+				mounts_mutex_unlock(ap->parent);
+				master_source_unlock(ap->parent->entry);
+			}
+
+			master_mutex_unlock();
+
+			pthread_setcancelstate(cur_state, NULL);
 
-			state_mutex_unlock(ap);
 		}
 	}
 
-	pthread_cleanup_pop(1);
-	sched_yield();
+	handle_mounts_cleanup(ap);
 
 	return NULL;
 }
@@ -1617,64 +1655,6 @@ static void key_thread_stdenv_vars_destroy(void *arg)
 	return;
 }
 
-static int is_automount_running(void)
-{
-	FILE *fp;
-	DIR *dir;
-	struct dirent entry;
-	struct dirent *result;
-	char path[PATH_MAX + 1], buf[PATH_MAX];
-
-	if ((dir = opendir("/proc")) == NULL) {
-		fprintf(stderr, "cannot opendir(/proc)\n");
-		exit(1);
-	}
-
-	while (readdir_r(dir, &entry, &result) == 0) {
-		int path_len, pid = 0;
-
-		if (!result)
-			break;
-
-		if (*entry.d_name == '.')
-			continue;
-
-		if (!strcmp(entry.d_name, "self"))
-			continue;
-
-		if (!isdigit(*entry.d_name))
-			continue;
-
-		pid = atoi(entry.d_name);
-		if (pid == getpid())
-			continue;
-
-		path_len = sprintf(path, "/proc/%s/cmdline", entry.d_name);
-		if (path_len >= PATH_MAX) {
-			fprintf(stderr,
-				"buffer to small for /proc path\n");
-			return -1;
-		}
-		path[path_len] = '\0';
-
-		fp = fopen(path, "r");
-		if (fp) {
-			int c, len = 0;
-
-			while (len < 127 && (c = fgetc(fp)) != EOF && c)
-				buf[len++] = c;
-			buf[len] = '\0';
-
-			if (strstr(buf, "automount"))
-				return pid;
-			fclose(fp);
-		}
-	}
-	closedir(dir);
-
-	return 0;
-}
-
 static void usage(void)
 {
 	fprintf(stderr,
@@ -1694,6 +1674,8 @@ static void usage(void)
 		"			specify global mount options\n"
 		"	-l --set-log-priority priority path [path,...]\n"
 		"			set daemon log verbosity\n"
+		"	-C --dont-check-daemon\n"
+		"			don't check if daemon is already running\n"
 		"	-V --version	print version, build config and exit\n"
 		, program);
 }
@@ -1768,6 +1750,26 @@ static void show_build_info(void)
 #ifdef WITH_DMALLOC
 	printf("WITH_DMALLOC ");
 	count = count + 13;
+
+	if (count > 60) {
+		printf("\n  ");
+		count = 0;
+	}
+#endif
+
+#ifdef LIBXML2_WORKAROUND
+	printf("LIBXML2_WORKAROUND ");
+	count = count + 19;
+
+	if (count > 60) {
+		printf("\n  ");
+		count = 0;
+	}
+#endif
+
+#ifdef WITH_LIBTIRPC
+	printf("WITH_LIBTIRPC ");
+	count = count + 14;
 #endif
 
 	printf("\n\n");
@@ -1814,11 +1816,10 @@ int main(int argc, char *argv[])
 {
 	int res, opt, status;
 	int logpri = -1;
-	unsigned ghost, logging;
+	unsigned ghost, logging, daemon_check;
 	unsigned foreground, have_global_options;
 	time_t timeout;
 	time_t age = time(NULL);
-	sigset_t allsigs;
 	struct rlimit rlim;
 	static const struct option long_options[] = {
 		{"help", 0, 0, 'h'},
@@ -1833,11 +1834,20 @@ int main(int argc, char *argv[])
 		{"global-options", 1, 0, 'O'},
 		{"version", 0, 0, 'V'},
 		{"set-log-priority", 1, 0, 'l'},
+		{"dont-check-daemon", 0, 0, 'C'},
+		{"force", 0, 0, 'F'},
 		{0, 0, 0, 0}
 	};
 
-	sigfillset(&allsigs);
-	sigprocmask(SIG_BLOCK, &allsigs, NULL);
+	sigfillset(&block_sigs);
+	/* allow for the dropping of core files */
+	sigdelset(&block_sigs, SIGABRT);
+	sigdelset(&block_sigs, SIGBUS);
+	sigdelset(&block_sigs, SIGSEGV);
+	sigdelset(&block_sigs, SIGILL);
+	sigdelset(&block_sigs, SIGFPE);
+	sigdelset(&block_sigs, SIGTRAP);
+	sigprocmask(SIG_BLOCK, &block_sigs, NULL);
 
 	program = argv[0];
 
@@ -1851,9 +1861,10 @@ int main(int argc, char *argv[])
 	global_options = NULL;
 	have_global_options = 0;
 	foreground = 0;
+	daemon_check = 1;
 
 	opterr = 0;
-	while ((opt = getopt_long(argc, argv, "+hp:t:vdD:fVrO:l:n:", long_options, NULL)) != EOF) {
+	while ((opt = getopt_long(argc, argv, "+hp:t:vdD:fVrO:l:n:CF", long_options, NULL)) != EOF) {
 		switch (opt) {
 		case 'h':
 			usage();
@@ -1922,6 +1933,14 @@ int main(int argc, char *argv[])
 			}
 			break;
 
+		case 'C':
+			daemon_check = 0;
+			break;
+
+		case 'F':
+			do_force_unlink = 1;
+			break;
+
 		case '?':
 		case ':':
 			printf("%s: Ambiguous or unknown options\n", program);
@@ -1965,11 +1984,6 @@ int main(int argc, char *argv[])
 		exit(exit_code);
 	}
 
-	if (is_automount_running() > 0) {
-		fprintf(stderr, "%s: program is already running.\n",
-			program);
-		exit(1);
-	}
 #if 0
 	if (!load_autofs4_module()) {
 		fprintf(stderr, "%s: can't load %s filesystem module.\n",
@@ -1980,7 +1994,9 @@ int main(int argc, char *argv[])
 
 	if (!query_kproto_ver() || get_kver_major() < 5) {
 		fprintf(stderr,
-			"%s: kernel protocol version 5.00 or above required.\n",
+			"%s: test mount forbidden or "
+			"incorrect kernel protocol version, "
+			"kernel protocol version 5.00 or above required.\n",
 			program);
 		exit(1);
 	}
@@ -2001,7 +2017,7 @@ int main(int argc, char *argv[])
 		     "can't increase core file limit - continuing");
 #endif
 
-	become_daemon(foreground);
+	become_daemon(foreground, daemon_check);
 
 	if (argc == 0)
 		master_list = master_new(NULL, timeout, ghost);
@@ -2011,31 +2027,48 @@ int main(int argc, char *argv[])
 	if (!master_list) {
 		logerr("%s: can't create master map %s",
 			program, argv[0]);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
 		exit(1);
 	}
 
-	if (pthread_attr_init(&thread_attr)) {
+	if (pthread_attr_init(&th_attr)) {
 		logerr("%s: failed to init thread attribute struct!",
 		     program);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
+		exit(1);
+	}
+
+	if (pthread_attr_init(&th_attr_detached)) {
+		logerr("%s: failed to init thread attribute struct!",
+		     program);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+		close(start_pipefd[1]);
+		release_flag_file();
 		exit(1);
 	}
 
 	if (pthread_attr_setdetachstate(
-			&thread_attr, PTHREAD_CREATE_DETACHED)) {
+			&th_attr_detached, PTHREAD_CREATE_DETACHED)) {
 		logerr("%s: failed to set detached thread attribute!",
 		     program);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
 		exit(1);
 	}
 
 #ifdef _POSIX_THREAD_ATTR_STACKSIZE
 	if (pthread_attr_setstacksize(
-			&thread_attr, PTHREAD_STACK_MIN*64)) {
+			&th_attr_detached, PTHREAD_STACK_MIN*64)) {
 		logerr("%s: failed to set stack size thread attribute!",
 		       program);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
 		exit(1);
 	}
 #endif
@@ -2051,26 +2084,43 @@ int main(int argc, char *argv[])
 		logerr("%s: failed to create thread data key for std env vars!",
 		       program);
 		master_kill(master_list);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
 		exit(1);
 	}
 
+	init_ioctl_ctl();
+
 	if (!alarm_start_handler()) {
 		logerr("%s: failed to create alarm handler thread!", program);
 		master_kill(master_list);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
 		exit(1);
 	}
 
 	if (!st_start_handler()) {
 		logerr("%s: failed to create FSM handler thread!", program);
 		master_kill(master_list);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
 		exit(1);
 	}
 
 #ifdef LIBXML2_WORKAROUND
-	void *dh = dlopen("libxml2.so", RTLD_NOW);
+	void *dh_xml2 = dlopen("libxml2.so", RTLD_NOW);
+	if (!dh_xml2)
+		dh_xml2 = dlopen("libxml2.so.2", RTLD_NOW);
+	if (dh_xml2)
+		xmlInitParser();
+#endif
+#ifdef TIRPC_WORKAROUND
+	void *dh_tirpc = dlopen("libitirpc.so", RTLD_NOW);
+	if (!dh_tirpc)
+		dh_tirpc = dlopen("libitirpc.so.1", RTLD_NOW);
 #endif
 
 	if (!master_read_master(master_list, age, 0)) {
@@ -2078,9 +2128,17 @@ int main(int argc, char *argv[])
 		*pst_stat = 3;
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
+		release_flag_file();
 		exit(3);
 	}
 
+	/*
+	 * Mmm ... reset force unlink umount so we don't also do this
+	 * in future when we receive a HUP signal.
+	 */
+	do_force_unlink = 0;
+
+	st_stat = 0;
 	res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 	close(start_pipefd[1]);
 
@@ -2094,11 +2152,20 @@ int main(int argc, char *argv[])
 		pid_file = NULL;
 	}
 	closelog();
+	release_flag_file();
 
-#ifdef LIBXML2_WORKAROUND
-	if (dh)
-		dlclose(dh);
+#ifdef TIRPC_WORKAROUND
+	if (dh_tirpc)
+		dlclose(dh_tirpc);
 #endif
+#ifdef LIBXML2_WORKAROUND
+	if (dh_xml2) {
+		xmlCleanupParser();
+		dlclose(dh_xml2);
+	}
+#endif
+	close_ioctl_ctl();
+
 	info(logging, "autofs stopped");
 
 	exit(0);
