@@ -19,22 +19,16 @@
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <sys/mount.h>
-#include <sys/wait.h>
-#include <ctype.h>
 #include <stdio.h>
 #include <dirent.h>
 #include <sys/vfs.h>
 #include <pwd.h>
 #include <grp.h>
-#include <libgen.h>
 
 #include "automount.h"
 
 #define MAX_OPTIONS_LEN		80
 #define MAX_MNT_NAME_LEN	30
-#define MAX_ENV_NAME		15
-
-#define EBUFSIZ 1024
 
 const unsigned int t_indirect = AUTOFS_TYPE_INDIRECT;
 const unsigned int t_direct = AUTOFS_TYPE_DIRECT;
@@ -48,50 +42,9 @@ static const char mnt_name_template[]      = "automount(pid%u)";
 static struct kernel_mod_version kver = {0, 0};
 static const char kver_options_template[]  = "fd=%d,pgrp=%u,minproto=3,maxproto=5";
 
-#define EXT_MOUNTS_HASH_SIZE    50
-
-struct ext_mount {
-	char *mountpoint;
-	unsigned int umount;
-	struct list_head mount;
-	struct list_head mounts;
-};
-static struct list_head ext_mounts_hash[EXT_MOUNTS_HASH_SIZE];
-static unsigned int ext_mounts_hash_init_done = 0;
-static pthread_mutex_t ext_mount_hash_mutex = PTHREAD_MUTEX_INITIALIZER;
-
-unsigned int linux_version_code(void)
-{
-	struct utsname my_utsname;
-	unsigned int p, q, r;
-	char *tmp, *save;
-
-	if (uname(&my_utsname))
-		return 0;
-
-	p = q = r = 0;
-
-	tmp = strtok_r(my_utsname.release, ".", &save);
-	if (!tmp)
-		return 0;
-	p = (unsigned int ) atoi(tmp);
-
-	tmp = strtok_r(NULL, ".", &save);
-	if (!tmp)
-		return KERNEL_VERSION(p, 0, 0);
-	q = (unsigned int) atoi(tmp);
-
-	tmp = strtok_r(NULL, ".", &save);
-	if (!tmp)
-		return KERNEL_VERSION(p, q, 0);
-	r = (unsigned int) atoi(tmp);
-
-	return KERNEL_VERSION(p, q, r);
-}
-
 unsigned int query_kproto_ver(void)
 {
-	struct ioctl_ops *ops;
+	struct ioctl_ops *ops = get_ioctl_ops();
 	char dir[] = "/tmp/autoXXXXXX", *t_dir;
 	char options[MAX_OPTIONS_LEN + 1];
 	pid_t pgrp = getpgrp();
@@ -132,19 +85,10 @@ unsigned int query_kproto_ver(void)
 		return 0;
 	}
 
-	ops = get_ioctl_ops();
-	if (!ops) {
-		umount(t_dir);
-		close(pipefd[0]);
-		rmdir(t_dir);
-		return 0;
-	}
-
 	ops->open(LOGOPT_NONE, &ioctlfd, st.st_dev, t_dir);
 	if (ioctlfd == -1) {
 		umount(t_dir);
 		close(pipefd[0]);
-		close_ioctl_ctl();
 		rmdir(t_dir);
 		return 0;
 	}
@@ -156,7 +100,6 @@ unsigned int query_kproto_ver(void)
 		ops->close(LOGOPT_NONE, ioctlfd);
 		umount(t_dir);
 		close(pipefd[0]);
-		close_ioctl_ctl();
 		rmdir(t_dir);
 		return 0;
 	}
@@ -166,7 +109,6 @@ unsigned int query_kproto_ver(void)
 		ops->close(LOGOPT_NONE, ioctlfd);
 		umount(t_dir);
 		close(pipefd[0]);
-		close_ioctl_ctl();
 		rmdir(t_dir);
 		return 0;
 	}
@@ -174,7 +116,6 @@ unsigned int query_kproto_ver(void)
 	ops->close(LOGOPT_NONE, ioctlfd);
 	umount(t_dir);
 	close(pipefd[0]);
-	close_ioctl_ctl();
 	rmdir(t_dir);
 
 	return 1;
@@ -190,408 +131,6 @@ unsigned int get_kver_minor(void)
 	return kver.minor;
 }
 
-#ifdef HAVE_MOUNT_NFS
-static int extract_version(char *start, struct nfs_mount_vers *vers)
-{
-	char *s_ver = strchr(start, ' ');
-	if (!s_ver)
-		return 0;
-	while (*s_ver && !isdigit(*s_ver)) {
-		s_ver++;
-		if (!*s_ver)
-			return 0;
-		break;
-	}
-	vers->major = atoi(strtok(s_ver, "."));
-	vers->minor = (unsigned int) atoi(strtok(NULL, "."));
-	vers->fix = (unsigned int) atoi(strtok(NULL, "."));
-	return 1;
-}
-
-int check_nfs_mount_version(struct nfs_mount_vers *vers,
-			    struct nfs_mount_vers *check)
-{
-	pid_t f;
-	int ret, status, pipefd[2];
-	char errbuf[EBUFSIZ + 1], *p, *sp;
-	int errp, errn;
-	sigset_t allsigs, tmpsig, oldsig;
-	char *s_ver;
-	int cancel_state;
-
-	if (pipe(pipefd))
-		return -1;
-
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel_state);
-
-	sigfillset(&allsigs);
-	pthread_sigmask(SIG_BLOCK, &allsigs, &oldsig);
-
-	f = fork();
-	if (f == 0) {
-		reset_signals();
-		close(pipefd[0]);
-		dup2(pipefd[1], STDOUT_FILENO);
-		dup2(pipefd[1], STDERR_FILENO);
-		close(pipefd[1]);
-
-		execl(PATH_MOUNT_NFS, PATH_MOUNT_NFS, "-V", (char *) NULL);
-		_exit(255);	/* execv() failed */
-	}
-
-	ret = 0;
-
-	tmpsig = oldsig;
-
-	sigaddset(&tmpsig, SIGCHLD);
-	pthread_sigmask(SIG_SETMASK, &tmpsig, NULL);
-
-	close(pipefd[1]);
-
-	if (f < 0) {
-		close(pipefd[0]);
-		pthread_sigmask(SIG_SETMASK, &oldsig, NULL);
-		pthread_setcancelstate(cancel_state, NULL);
-		return -1;
-	}
-
-	errp = 0;
-	do {
-		while (1) {
-			errn = read(pipefd[0], errbuf + errp, EBUFSIZ - errp);
-			if (errn == -1 && errno == EINTR)
-				continue;
-			break;
-		}
-
-		if (errn > 0) {
-			errp += errn;
-
-			sp = errbuf;
-			while (errp && (p = memchr(sp, '\n', errp))) {
-				*p++ = '\0';
-				errp -= (p - sp);
-				sp = p;
-			}
-
-			if (errp && sp != errbuf)
-				memmove(errbuf, sp, errp);
-
-			if (errp >= EBUFSIZ) {
-				/* Line too long, split */
-				errbuf[errp] = '\0';
-				if ((s_ver = strstr(errbuf, "nfs-utils"))) {
-					if (extract_version(s_ver, vers))
-						ret = 1;
-				}
-				errp = 0;
-			}
-
-			if ((s_ver = strstr(errbuf, "nfs-utils"))) {
-				if (extract_version(s_ver, vers))
-					ret = 1;
-			}
-		}
-	} while (errn > 0);
-
-	close(pipefd[0]);
-
-	if (errp > 0) {
-		/* End of file without \n */
-		errbuf[errp] = '\0';
-		if ((s_ver = strstr(errbuf, "nfs-utils"))) {
-			if (extract_version(s_ver, vers))
-				ret = 1;
-		}
-	}
-
-	if (ret) {
-		if ((vers->major < check->major) ||
-		    ((vers->major == check->major) && (vers->minor < check->minor)) ||
-		    ((vers->major == check->major) && (vers->minor == check->minor) &&
-		     (vers->fix < check->fix)))
-			ret = 0;
-	}
-
-	if (waitpid(f, &status, 0) != f)
-		debug(LOGOPT_NONE, "no process found to wait for");
-
-	pthread_sigmask(SIG_SETMASK, &oldsig, NULL);
-	pthread_setcancelstate(cancel_state, NULL);
-
-	return ret;
-}
-#else
-int check_nfs_mount_version(struct nfs_mount_vers *vers,
-			    struct nfs_mount_vers *check)
-{
-	return 0;
-}
-#endif
-
-static char *set_env_name(const char *prefix, const char *name, char *buf)
-{
-	size_t len;
-
-	len = strlen(name);
-	if (prefix)
-		len += strlen(prefix);
-	len++;
-
-	if (len > MAX_ENV_NAME)
-		return NULL;
-
-	if (!prefix)
-		strcpy(buf, name);
-	else {
-		strcpy(buf, prefix);
-		strcat(buf, name);
-	}
-	return buf;
-}
-
-static struct substvar *do_macro_addvar(struct substvar *list,
-					const char *prefix,
-					const char *name,
-					const char *val)
-{
-	char buf[MAX_ENV_NAME + 1];
-	char *new;
-	size_t len;
-
-	new = set_env_name(prefix, name, buf);
-	if (new) {
-		len = strlen(new);
-		list = macro_addvar(list, new, len, val);
-	}
-	return list;
-}
-
-static struct substvar *do_macro_removevar(struct substvar *list,
-					   const char *prefix,
-					   const char *name)
-{
-	char buf[MAX_ENV_NAME + 1];
-	char *new;
-	size_t len;
-
-	new = set_env_name(prefix, name, buf);
-	if (new) {
-		len = strlen(new);
-		list = macro_removevar(list, new, len);
-	}
-	return list;
-}
-
-struct substvar *addstdenv(struct substvar *sv, const char *prefix)
-{
-	struct substvar *list = sv;
-	struct thread_stdenv_vars *tsv;
-	char numbuf[16];
-
-	tsv = pthread_getspecific(key_thread_stdenv_vars);
-	if (tsv) {
-		const struct substvar *mv;
-		int ret;
-		long num;
-
-		num = (long) tsv->uid;
-		ret = sprintf(numbuf, "%ld", num);
-		if (ret > 0)
-			list = do_macro_addvar(list, prefix, "UID", numbuf);
-		num = (long) tsv->gid;
-		ret = sprintf(numbuf, "%ld", num);
-		if (ret > 0)
-			list = do_macro_addvar(list, prefix, "GID", numbuf);
-		list = do_macro_addvar(list, prefix, "USER", tsv->user);
-		list = do_macro_addvar(list, prefix, "GROUP", tsv->group);
-		list = do_macro_addvar(list, prefix, "HOME", tsv->home);
-		mv = macro_findvar(list, "HOST", 4);
-		if (mv) {
-			char *shost = strdup(mv->val);
-			if (shost) {
-				char *dot = strchr(shost, '.');
-				if (dot)
-					*dot = '\0';
-				list = do_macro_addvar(list,
-						       prefix, "SHOST", shost);
-				free(shost);
-			}
-		}
-	}
-	return list;
-}
-
-struct substvar *removestdenv(struct substvar *sv, const char *prefix)
-{
-	struct substvar *list = sv;
-
-	list = do_macro_removevar(list, prefix, "UID");
-	list = do_macro_removevar(list, prefix, "USER");
-	list = do_macro_removevar(list, prefix, "HOME");
-	list = do_macro_removevar(list, prefix, "GID");
-	list = do_macro_removevar(list, prefix, "GROUP");
-	list = do_macro_removevar(list, prefix, "SHOST");
-	return list;
-}
-
-void add_std_amd_vars(struct substvar *sv)
-{
-	char *tmp;
-
-	tmp = conf_amd_get_arch();
-	if (tmp) {
-		macro_global_addvar("arch", 4, tmp);
-		free(tmp);
-	}
-
-	tmp = conf_amd_get_karch();
-	if (tmp) {
-		macro_global_addvar("karch", 5, tmp);
-		free(tmp);
-	}
-
-	tmp = conf_amd_get_os();
-	if (tmp) {
-		macro_global_addvar("os", 2, tmp);
-		free(tmp);
-	}
-
-	tmp = conf_amd_get_full_os();
-	if (tmp) {
-		macro_global_addvar("full_os", 7, tmp);
-		free(tmp);
-	}
-
-	tmp = conf_amd_get_os_ver();
-	if (tmp) {
-		macro_global_addvar("osver", 5, tmp);
-		free(tmp);
-	}
-
-	tmp = conf_amd_get_vendor();
-	if (tmp) {
-		macro_global_addvar("vendor", 6, tmp);
-		free(tmp);
-	}
-
-	/* Umm ... HP_UX cluster name, probably not used */
-	tmp = conf_amd_get_cluster();
-	if (tmp) {
-		macro_global_addvar("cluster", 7, tmp);
-		free(tmp);
-	} else {
-		const struct substvar *v = macro_findvar(sv, "domain", 4);
-		if (v && *v->val) {
-			tmp = strdup(v->val);
-			if (tmp)
-				macro_global_addvar("cluster", 7, tmp);
-		}
-	}
-
-	tmp = conf_amd_get_auto_dir();
-	if (tmp) {
-		macro_global_addvar("autodir", 7, tmp);
-		free(tmp);
-	}
-
-	return;
-}
-
-void remove_std_amd_vars(void)
-{
-	macro_global_removevar("autodir", 7);
-	macro_global_removevar("cluster", 7);
-	macro_global_removevar("vendor", 6);
-	macro_global_removevar("osver", 5);
-	macro_global_removevar("full_os", 7);
-	macro_global_removevar("os", 2);
-	macro_global_removevar("karch", 5);
-	macro_global_removevar("arch", 4);
-	return;
- }
-
-struct amd_entry *new_amd_entry(const struct substvar *sv)
-{
-	struct amd_entry *new;
-	const struct substvar *v;
-	char *path;
-
-	v = macro_findvar(sv, "path", 4);
-	if (!v)
-		return NULL;
-
-	path = strdup(v->val);
-	if (!path)
-		return NULL;
-
-	new = malloc(sizeof(struct amd_entry));
-	if (!new) {
-		free(path);
-		return NULL;
-	}
-
-	memset(new, 0, sizeof(*new));
-	new->path = path;
-	INIT_LIST_HEAD(&new->list);
-	INIT_LIST_HEAD(&new->entries);
-	INIT_LIST_HEAD(&new->ext_mount);
-
-	return new;
-}
-
-void clear_amd_entry(struct amd_entry *entry)
-{
-	if (!entry)
-		return;
-	if (entry->path)
-		free(entry->path);
-	if (entry->map_type)
-		free(entry->map_type);
-	if (entry->pref)
-		free(entry->pref);
-	if (entry->fs)
-		free(entry->fs);
-	if (entry->rhost)
-		free(entry->rhost);
-	if (entry->rfs)
-		free(entry->rfs);
-	if (entry->opts)
-		free(entry->opts);
-	if (entry->addopts)
-		free(entry->addopts);
-	if (entry->remopts)
-		free(entry->remopts);
-	if (entry->sublink)
-		free(entry->sublink);
-	if (entry->selector)
-		free_selector(entry->selector);
-	return;
-}
-
-void free_amd_entry(struct amd_entry *entry)
-{
-	clear_amd_entry(entry);
-	free(entry);
-	return;
-}
-
-void free_amd_entry_list(struct list_head *entries)
-{
-	if (!list_empty(entries)) {
-		struct list_head *head = entries;
-		struct amd_entry *this;
-		struct list_head *p;
-
-		p = head->next;
-		while (p != head) {
-			this = list_entry(p, struct amd_entry, list);
-			p = p->next;
-			free_amd_entry(this);
-		}
-	}
-}
-
 /*
  * Make common autofs mount options string
  */
@@ -660,129 +199,6 @@ char *make_mnt_name_string(char *path)
 	return mnt_name;
 }
 
-static void ext_mounts_hash_init(void)
-{
-	int i;
-	for (i = 0; i < EXT_MOUNTS_HASH_SIZE; i++)
-		INIT_LIST_HEAD(&ext_mounts_hash[i]);
-	ext_mounts_hash_init_done = 1;
-}
-
-static struct ext_mount *ext_mount_lookup(const char *mountpoint)
-{
-	u_int32_t hval = hash(mountpoint, EXT_MOUNTS_HASH_SIZE);
-	struct list_head *p, *head;
-
-	if (!ext_mounts_hash_init_done)
-		ext_mounts_hash_init();
-
-	if (list_empty(&ext_mounts_hash[hval]))
-		return NULL;
-
-	head = &ext_mounts_hash[hval];
-	list_for_each(p, head) {
-		struct ext_mount *this = list_entry(p, struct ext_mount, mount);
-		if (!strcmp(this->mountpoint, mountpoint))
-			return this;
-	}
-	return NULL;
-}
-
-int ext_mount_add(struct list_head *entry, const char *path, unsigned int umount)
-{
-	struct ext_mount *em;
-	char *auto_dir;
-	u_int32_t hval;
-	int ret = 0;
-
-	/* Not a mount in the external mount directory */
-	auto_dir = conf_amd_get_auto_dir();
-	if (strncmp(path, auto_dir, strlen(auto_dir))) {
-		free(auto_dir);
-		return 0;
-	}
-	free(auto_dir);
-
-	pthread_mutex_lock(&ext_mount_hash_mutex);
-
-	em = ext_mount_lookup(path);
-	if (em) {
-		struct list_head *p, *head;
-		head = &em->mounts;
-		list_for_each(p, head) {
-			if (p == entry)
-				goto done;
-		}
-		list_add_tail(entry, &em->mounts);
-		ret = 1;
-		goto done;
-	}
-
-	em = malloc(sizeof(struct ext_mount));
-	if (!em) {
-		ret = -1;
-		goto done;
-	}
-
-	em->mountpoint = strdup(path);
-	if (!em->mountpoint) {
-		free(em);
-		ret = -1;
-		goto done;
-	}
-	em->umount = umount;
-	INIT_LIST_HEAD(&em->mount);
-	INIT_LIST_HEAD(&em->mounts);
-
-	hval = hash(path, EXT_MOUNTS_HASH_SIZE);
-	list_add_tail(&em->mount, &ext_mounts_hash[hval]);
-
-	list_add_tail(entry, &em->mounts);
-
-	ret = 1;
-done:
-	pthread_mutex_unlock(&ext_mount_hash_mutex);
-	return ret;
-}
-
-int ext_mount_remove(struct list_head *entry, const char *path)
-{
-	struct ext_mount *em;
-	char *auto_dir;
-	int ret = 0;
-
-	/* Not a mount in the external mount directory */
-	auto_dir = conf_amd_get_auto_dir();
-	if (strncmp(path, auto_dir, strlen(auto_dir))) {
-		free(auto_dir);
-		return 0;
-	}
-	free(auto_dir);
-
-	pthread_mutex_lock(&ext_mount_hash_mutex);
-
-	em = ext_mount_lookup(path);
-	if (!em)
-		goto done;
-
-	list_del_init(entry);
-
-	if (!list_empty(&em->mounts))
-		goto done;
-	else {
-		list_del_init(&em->mount);
-		if (em->umount)
-			ret = 1;
-		if (list_empty(&em->mount)) {
-			free(em->mountpoint);
-			free(em);
-		}
-	}
-done:
-	pthread_mutex_unlock(&ext_mount_hash_mutex);
-	return ret;
-}
-
 /*
  * Get list of mounts under path in longest->shortest order
  */
@@ -1090,6 +506,100 @@ int has_fstab_option(const char *opt)
 	return ret;
 }
 
+char *get_offset(const char *prefix, char *offset,
+		 struct list_head *head, struct list_head **pos)
+{
+	struct list_head *next;
+	struct mnt_list *this;
+	size_t plen = strlen(prefix);
+	size_t len = 0;
+
+	*offset = '\0';
+	next = *pos ? (*pos)->next : head->next;
+	while (next != head) {
+		char *pstart, *pend;
+
+		this = list_entry(next, struct mnt_list, ordered);
+		*pos = next;
+		next = next->next;
+
+		if (strlen(this->path) <= plen)
+			continue;
+
+		if (!strncmp(prefix, this->path, plen)) {
+			pstart = &this->path[plen];
+
+			/* not part of this sub-tree */
+			if (*pstart != '/')
+				continue;
+
+			/* get next offset */
+			pend = pstart;
+			while (*pend++) ;
+			len = pend - pstart - 1;
+			strncpy(offset, pstart, len);
+			offset[len] ='\0';
+			break;
+		}
+	}
+
+	while (next != head) {
+		char *pstart;
+
+		this = list_entry(next, struct mnt_list, ordered);
+
+		if (strlen(this->path) <= plen + len)
+			break;
+
+		pstart = &this->path[plen];
+
+		/* not part of this sub-tree */
+		if (*pstart != '/')
+			break;
+
+		/* new offset */
+		if (!*(pstart + len + 1))
+			break;
+
+		/* compare next offset */
+		if (pstart[len] != '/' || strncmp(offset, pstart, len))
+			break;
+
+		*pos = next;
+		next = next->next;
+	}
+
+	return *offset ? offset : NULL;
+}
+
+void add_ordered_list(struct mnt_list *ent, struct list_head *head)
+{
+	struct list_head *p;
+	struct mnt_list *this;
+
+	list_for_each(p, head) {
+		size_t tlen;
+		int eq;
+
+		this = list_entry(p, struct mnt_list, ordered);
+		tlen = strlen(this->path);
+
+		eq = strncmp(this->path, ent->path, tlen);
+		if (!eq && tlen == strlen(ent->path))
+			return;
+
+		if (eq > 0) {
+			INIT_LIST_HEAD(&ent->ordered);
+			list_add_tail(&ent->ordered, p);
+			return;
+		}
+	}
+	INIT_LIST_HEAD(&ent->ordered);
+	list_add_tail(&ent->ordered, p);
+
+	return;
+}
+
 /*
  * Since we have to look at the entire mount tree for direct
  * mounts (all mounts under "/") and we may have a large number
@@ -1189,6 +699,7 @@ struct mnt_list *tree_make_mnt_tree(const char *table, const char *path)
 		INIT_LIST_HEAD(&ent->list);
 		INIT_LIST_HEAD(&ent->entries);
 		INIT_LIST_HEAD(&ent->sublist);
+		INIT_LIST_HEAD(&ent->ordered);
 
 		ent->path = malloc(len + 1);
 		if (!ent->path) {
@@ -1542,7 +1053,7 @@ void set_tsd_user_vars(unsigned int logopt, uid_t uid, gid_t gid)
 	/* Try to get group info */
 
 	grplen = sysconf(_SC_GETGR_R_SIZE_MAX);
-	if (grplen < 0) {
+	if (tmplen < 0) {
 		error(logopt, "failed to get buffer size for getgrgid_r");
 		goto free_tsv_home;
 	}
@@ -1617,12 +1128,13 @@ const char *mount_type_str(const unsigned int type)
 }
 
 void notify_mount_result(struct autofs_point *ap,
-			 const char *path, time_t timeout, const char *type)
+			 const char *path, const char *type)
 {
-	if (timeout)
+	if (ap->exp_timeout)
 		info(ap->logopt,
 		    "mounted %s on %s with timeout %u, freq %u seconds",
-		    type, path, (unsigned int) timeout,
+		    type, path,
+		    (unsigned int) ap->exp_timeout,
 		    (unsigned int) ap->exp_runfreq);
 	else
 		info(ap->logopt,
@@ -1696,10 +1208,6 @@ static int do_remount_indirect(struct autofs_point *ap, int fd, const char *path
 			int i, j;
 
 			i = j = scandir(buf, &de2, 0, alphasort);
-			if (i < 0) {
-				free(de[n]);
-				continue;
-			}
 			while (i--)
 				free(de2[i]);
 			free(de2);
@@ -1730,14 +1238,16 @@ static int do_remount_indirect(struct autofs_point *ap, int fd, const char *path
 }
 
 static int remount_active_mount(struct autofs_point *ap,
-				struct mapent *me, const char *path, dev_t devid,
-				const unsigned int type, int *ioctlfd)
+				struct mapent_cache *mc,
+				const char *path, dev_t devid,
+				const unsigned int type,
+				int *ioctlfd)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
+	time_t timeout = ap->exp_timeout;
 	const char *str_type = mount_type_str(type);
 	char buf[MAX_ERR_BUF];
 	unsigned int mounted;
-	time_t timeout;
 	struct stat st;
 	int fd;
 
@@ -1747,17 +1257,10 @@ static int remount_active_mount(struct autofs_point *ap,
 	ops->open(ap->logopt, &fd, devid, path);
 	if (fd == -1)
 		return REMOUNT_OPEN_FAIL;
-	else {
-		if (type == t_indirect || type == t_offset)
-			timeout = ap->entry->maps->exp_timeout;
-		else
-			timeout = me->source->exp_timeout;
-	}
 
 	/* Re-reading the map, set timeout and return */
 	if (ap->state == ST_READMAP) {
-		debug(ap->logopt, "already mounted, update timeout");
-		ops->timeout(ap->logopt, fd, timeout);
+		ops->timeout(ap->logopt, fd, &timeout);
 		ops->close(ap->logopt, fd);
 		return REMOUNT_READ_MAP;
 	}
@@ -1779,7 +1282,7 @@ static int remount_active_mount(struct autofs_point *ap,
 		ops->close(ap->logopt, fd);
 		return REMOUNT_OPEN_FAIL;
 	}
-	ops->timeout(ap->logopt, fd, timeout);
+	ops->timeout(ap->logopt, fd, &timeout);
 	if (fstat(fd, &st) == -1) {
 		error(ap->logopt,
 		      "failed to stat %s mount %s", str_type, path);
@@ -1787,11 +1290,11 @@ static int remount_active_mount(struct autofs_point *ap,
 		ops->close(ap->logopt, fd);
 		return REMOUNT_STAT_FAIL;
 	}
-	if (type != t_indirect)
-		cache_set_ino_index(me->mc, path, st.st_dev, st.st_ino);
+	if (mc)
+		cache_set_ino_index(mc, path, st.st_dev, st.st_ino);
 	else
 		ap->dev = st.st_dev;
-	notify_mount_result(ap, path, timeout, str_type);
+	notify_mount_result(ap, path, str_type);
 
 	*ioctlfd = fd;
 
@@ -1834,20 +1337,24 @@ static int remount_active_mount(struct autofs_point *ap,
 int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
+	struct mapent_cache *mc;
 	const char *path;
 	int ret, fd;
 	dev_t devid;
 
-	if (type == t_indirect)
+	if (type == t_indirect) {
+		mc = NULL;
 		path = ap->path;
-	else
+	} else {
+		mc = me->mc;
 		path = me->key;
+	}
 
 	ret = ops->mount_device(ap->logopt, path, type, &devid);
 	if (ret == -1 || ret == 0)
 		return -1;
 
-	ret = remount_active_mount(ap, me, path, devid, type, &fd);
+	ret = remount_active_mount(ap, mc, path, devid, type, &fd);
 
 	/*
 	 * The directory must exist since we found a device
@@ -1871,9 +1378,7 @@ int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 	 * record that in the mount point struct. Otherwise we're
 	 * re-reading the map.
 	*/
-	if (ret == REMOUNT_READ_MAP)
-		return 1;
-	else if (ret == REMOUNT_SUCCESS) {
+	if (ret == REMOUNT_SUCCESS || ret == REMOUNT_READ_MAP) {
 		if (fd != -1) {
 			if (type == t_indirect)
 				ap->ioctlfd = fd;
@@ -1928,33 +1433,6 @@ int umount_ent(struct autofs_point *ap, const char *path)
 	return rv;
 }
 
-static int do_mount_autofs_offset(struct autofs_point *ap,
-				  struct mapent *oe, const char *root,
-				  char *offset)
-
-{
-	int mounted = 0;
-	int ret;
-
-	debug(ap->logopt, "mount offset %s at %s", oe->key, root);
-
-	ret = mount_autofs_offset(ap, oe, root, offset);
-	if (ret >= MOUNT_OFFSET_OK)
-		mounted++;
-	else {
-		if (ret != MOUNT_OFFSET_IGNORE)
-			warn(ap->logopt, "failed to mount offset");
-		else {
-			debug(ap->logopt, "ignoring \"nohide\" trigger %s",
-			      oe->key);
-			free(oe->mapent);
-			oe->mapent = NULL;
-		}
-	}
-
-	return mounted;
-}
-
 int mount_multi_triggers(struct autofs_point *ap, struct mapent *me,
 			 const char *root, unsigned int start, const char *base)
 {
@@ -1963,7 +1441,8 @@ int mount_multi_triggers(struct autofs_point *ap, struct mapent *me,
 	struct mapent *oe;
 	struct list_head *pos = NULL;
 	unsigned int fs_path_len;
-	int mounted;
+	unsigned int mounted;
+	int ret;
 
 	fs_path_len = start + strlen(base);
 	if (fs_path_len > PATH_MAX)
@@ -1983,20 +1462,20 @@ int mount_multi_triggers(struct autofs_point *ap, struct mapent *me,
 		if (!oe || !oe->mapent)
 			goto cont;
 
-		mounted += do_mount_autofs_offset(ap, oe, root, offset);
+		debug(ap->logopt, "mount offset %s at %s", oe->key, root);
 
-		/*
-		 * If re-constructing a multi-mount it's necessary to walk
-		 * into nested mounts, unlike the usual "mount only what's
-		 * needed as you go" behavior.
-		 */
-		if (ap->state == ST_READMAP && ap->flags & MOUNT_FLAG_REMOUNT) {
-			if (oe->ioctlfd != -1 ||
-			    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
-				char oe_root[PATH_MAX + 1];
-				strcpy(oe_root, root);
-				strcat(oe_root, offset); 
-				mount_multi_triggers(ap, oe, oe_root, strlen(oe_root), base);
+		ret = mount_autofs_offset(ap, oe, root, offset);
+		if (ret >= MOUNT_OFFSET_OK)
+			mounted++;
+		else {
+			if (ret != MOUNT_OFFSET_IGNORE)
+				warn(ap->logopt, "failed to mount offset");
+			else {
+				debug(ap->logopt,
+				      "ignoring \"nohide\" trigger %s",
+				      oe->key);
+				free(oe->mapent);
+				oe->mapent = NULL;
 			}
 		}
 cont:
@@ -2007,41 +1486,6 @@ cont:
 	return mounted;
 }
 
-static int rmdir_path_offset(struct autofs_point *ap, struct mapent *oe)
-{
-	char *dir, *path;
-	unsigned int split;
-	int ret;
-
-	if (ap->type == LKP_DIRECT)
-		return rmdir_path(ap, oe->key, oe->multi->dev);
-
-	dir = strdup(oe->key);
-
-	if (ap->flags & MOUNT_FLAG_GHOST)
-		split = strlen(ap->path) + strlen(oe->multi->key) + 1;
-	else
-		split = strlen(ap->path);
-
-	dir[split] = '\0';
-	path = &dir[split + 1];
-
-	if (chdir(dir) == -1) {
-		error(ap->logopt, "failed to chdir to %s", dir);
-		free(dir);
-		return -1;
-	}
-
-	ret = rmdir_path(ap, path, ap->dev);
-
-	free(dir);
-
-	if (chdir("/") == -1)
-		error(ap->logopt, "failed to chdir to /");
-
-	return ret;
-}
-
 int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root, const char *base)
 {
 	char path[PATH_MAX + 1];
@@ -2065,12 +1509,13 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 	pos = NULL;
 	offset = path;
 
+	/* Make sure "none" of the offsets have an active mount. */
 	while ((offset = cache_get_offset(mm_base, offset, start, mm_root, &pos))) {
 		char *oe_base;
 
 		oe = cache_lookup_offset(mm_base, offset, start, &me->multi_list);
 		/* root offset is a special case */
-		if (!oe || (strlen(oe->key) - start) == 1)
+		if (!oe || !oe->mapent || (strlen(oe->key) - start) == 1)
 			continue;
 
 		/*
@@ -2080,37 +1525,28 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 		oe_base = oe->key + strlen(root);
 		left += umount_multi_triggers(ap, oe, root, oe_base);
 
-		if (oe->ioctlfd != -1 ||
-		    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
+		if (oe->ioctlfd != -1)
 			left++;
+	}
+
+	if (left)
+		return left;
+
+	pos = NULL;
+	offset = path;
+
+	/* Make sure "none" of the offsets have an active mount. */
+	while ((offset = cache_get_offset(mm_base, offset, start, mm_root, &pos))) {
+		oe = cache_lookup_offset(mm_base, offset, start, &me->multi_list);
+		/* root offset is a special case */
+		if (!oe || !oe->mapent || (strlen(oe->key) - start) == 1)
 			continue;
-		}
 
 		debug(ap->logopt, "umount offset %s", oe->key);
 
 		if (umount_autofs_offset(ap, oe)) {
 			warn(ap->logopt, "failed to umount offset");
 			left++;
-		} else {
-			struct stat st;
-			int ret;
-
-			if (!(oe->flags & MOUNT_FLAG_DIR_CREATED))
-				continue;
-
-			/*
-			 * An error due to partial directory removal is
-			 * ok so only try and remount the offset if the
-			 * actual mount point still exists.
-			 */
-			ret = rmdir_path_offset(ap, oe);
-			if (ret == -1 && !stat(oe->key, &st)) {
-				ret = do_mount_autofs_offset(ap, oe, root, offset);
-				if (ret)
-					left++;
-				/* But we did origianlly create this */
-				oe->flags |= MOUNT_FLAG_DIR_CREATED;
-			}
 		}
 	}
 
@@ -2126,8 +1562,7 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 		 */
 		if (is_mounted(_PATH_MOUNTED, root, MNTS_REAL)) {
 			info(ap->logopt, "unmounting dir = %s", root);
-			if (umount_ent(ap, root) &&
-			    is_mounted(_PATH_MOUNTED, root, MNTS_REAL)) {
+			if (umount_ent(ap, root)) {
 				if (mount_multi_triggers(ap, me, root, strlen(root), "/") < 0)
 					warn(ap->logopt,
 					     "failed to remount offset triggers");
@@ -2144,163 +1579,3 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 	return left;
 }
 
-int clean_stale_multi_triggers(struct autofs_point *ap,
-			       struct mapent *me, char *top, const char *base)
-{
-	char *root;
-	char mm_top[PATH_MAX + 1];
-	char path[PATH_MAX + 1];
-	char buf[MAX_ERR_BUF];
-	char *offset;
-	struct mapent *oe;
-	struct list_head *mm_root, *pos;
-	const char o_root[] = "/";
-	const char *mm_base;
-	int left, start;
-	time_t age;
-
-	if (top)
-		root = top;
-	else {
-		if (!strchr(me->multi->key, '/'))
-			/* Indirect multi-mount root */
-			/* sprintf okay - if it's mounted, it's
-			 * PATH_MAX or less bytes */
-			sprintf(mm_top, "%s/%s", ap->path, me->multi->key);
-		else
-			strcpy(mm_top, me->multi->key);
-		root = mm_top;
-	}
-
-	left = 0;
-	start = strlen(root);
-
-	mm_root = &me->multi->multi_list;
-
-	if (!base)
-		mm_base = o_root;
-	else
-		mm_base = base;
-
-	pos = NULL;
-	offset = path;
-	age = me->multi->age;
-
-	while ((offset = cache_get_offset(mm_base, offset, start, mm_root, &pos))) {
-		char *oe_base;
-		char *key;
-		int ret;
-
-		oe = cache_lookup_offset(mm_base, offset, start, &me->multi_list);
-		/* root offset is a special case */
-		if (!oe || (strlen(oe->key) - start) == 1)
-			continue;
-
-		/* Check for and umount stale subtree offsets */
-		oe_base = oe->key + strlen(root);
-		ret = clean_stale_multi_triggers(ap, oe, root, oe_base);
-		left += ret;
-		if (ret)
-			continue;
-
-		if (oe->age == age)
-			continue;
-
-		/*
-		 * If an offset that has an active mount has been removed
-		 * from the multi-mount we don't want to attempt to trigger
-		 * mounts for it. Obviously this is because it has been
-		 * removed, but less obvious is the potential strange
-		 * behaviour that can result if we do try and mount it
-		 * again after it's been expired. For example, if an NFS
-		 * file system is no longer exported and is later umounted
-		 * it can be mounted again without any error message but
-		 * shows as an empty directory. That's going to confuse
-		 * people for sure.
-		 *
-		 * If the mount cannot be umounted (the process is now
-		 * using a stale mount) the offset needs to be invalidated
-		 * so no further mounts will be attempted but the offset
-		 * cache entry must remain so expires can continue to
-		 * attempt to umount it. If the mount can be umounted and
-		 * the offset is removed, at least for NFS we will get
-		 * ESTALE errors when attempting list the directory.
-		 */
-		if (oe->ioctlfd != -1 ||
-		    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
-			if (umount_ent(ap, oe->key) &&
-			    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
-				debug(ap->logopt,
-				      "offset %s has active mount, invalidate",
-				      oe->key);
-				if (oe->mapent) {
-					free(oe->mapent);
-					oe->mapent = NULL;
-				}
-				left++;
-				continue;
-			}
-		}
-
-		key = strdup(oe->key);
-		if (!key) {
-	                char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		        error(ap->logopt, "malloc: %s", estr);
-			left++;
-			continue;
-		}
-
-		debug(ap->logopt, "umount offset %s", oe->key);
-
-		if (umount_autofs_offset(ap, oe)) {
-			warn(ap->logopt, "failed to umount offset %s", key);
-			left++;
-		} else {
-			struct stat st;
-
-			/* Mount point not ours to delete ? */
-			if (!(oe->flags & MOUNT_FLAG_DIR_CREATED)) {
-				debug(ap->logopt, "delete offset key %s", key);
-				if (cache_delete_offset(oe->mc, key) == CHE_FAIL)
-					error(ap->logopt,
-					     "failed to delete offset key %s", key);
-				free(key);
-				continue;
-			}
-
-			/*
-			 * An error due to partial directory removal is
-			 * ok so only try and remount the offset if the
-			 * actual mount point still exists.
-			 */
-			ret = rmdir_path_offset(ap, oe);
-			if (ret == -1 && !stat(oe->key, &st)) {
-				ret = do_mount_autofs_offset(ap, oe, root, offset);
-				if (ret) {
-					left++;
-					/* But we did origianlly create this */
-					oe->flags |= MOUNT_FLAG_DIR_CREATED;
-					free(key);
-					continue;
-				}
-				/*
-				 * Fall through if the trigger can't be mounted
-				 * again, since there is no offset there can't
-				 * be any mount requests so remove the map
-				 * entry from the cache. There's now a dead
-				 * offset mount, but what else can we do ....
-				 */
-			}
-
-			debug(ap->logopt, "delete offset key %s", key);
-
-			if (cache_delete_offset(oe->mc, key) == CHE_FAIL)
-				error(ap->logopt,
-				     "failed to delete offset key %s", key);
-		}
-		free(key);
-	}
-
-	return left;
-}
-
